// ============================================================================
// ==                                                                        ==
// == AMDiS - Adaptive multidimensional simulations                          ==
// ==                                                                        ==
// ==  http://www.amdis-fem.org                                              ==
// ==                                                                        ==
// ============================================================================
//
// Software License for AMDiS
//
// Copyright (c) 2010 Dresden University of Technology 
// All rights reserved.
// Authors: Simon Vey, Thomas Witkowski et al.
//
// This file is part of AMDiS
//
// See also license.opensource.txt in the distribution.



/** \file PetscSolverNavierStokes.h */

#ifndef AMDIS_PETSC_SOLVER_NAVIER_STOKES_H
#define AMDIS_PETSC_SOLVER_NAVIER_STOKES_H

#include "parallel/PetscSolverGlobalMatrix.h"

namespace AMDiS {

  using namespace std;

  struct NavierStokesSchurData {
    KSP kspMass;

    KSP kspLaplace;
    
    Mat matConDif;
  };

  class PetscSolverNavierStokes : public PetscSolverGlobalMatrix
  {
  private:    
    class IdFct : public AbstractFunction<double, double>
    {
    public:
      IdFct() 
	: AbstractFunction<double, double>(1)
      {}
      
      double operator()(const double& x) const 
      {
	return x;
      }
    };

    class MultConstFct : public AbstractFunction<double, double>
    {
    public:
      MultConstFct(double c) 
	: AbstractFunction<double, double>(1),
	  mConst(c)
      {}
      
      double operator()(const double& x) const 
      {
	return mConst * x;
      }
    private:
      double mConst;
    };

    class Multiplier3 : public BinaryAbstractFunction<double, double, double>
    {
    public:
      Multiplier3() 
	: BinaryAbstractFunction<double, double, double >(6) 
      {}
      
      double operator()(const double& phi, const double& phase) const 
      {
   	return  phase * phi;
      }
    };


    class EinsMinus : public AbstractFunction<double, double>
    {
    public:
      EinsMinus(double d)   
	: AbstractFunction<double, double>(2),
	  c(d)
      {}
      
      double operator()(const double& x) const 
      {
	return c * std::max(1.0-x,0.000001);
      }
    private:
      double c;
    };

    struct Multiplication : public BinaryAbstractFunction<double, double, double>{
      double operator()(const double &v1, const double &v2) const{
	return v2 / v1;
      }
    };
    


  public:
    /// Creator class
    class Creator : public OEMSolverCreator
    {
    public:
      virtual ~Creator() {}

      /// Returns a new PetscSolver object.
      OEMSolver* create() 
      { 
	return new PetscSolverNavierStokes(this->name); 
      }
    };

    PetscSolverNavierStokes(string name);

    void solvePetscMatrix(SystemVector &vec, AdaptInfo *adaptInfo);

    void setStokesData(double *nuPtr, double *invTauPtr, SystemVector *vec)
    {
      nu = nuPtr;
      invTau = invTauPtr;
      solution = vec;
    }
    
    void setPhase(DOFVector<double> *d)
    {
      phase = d;
    }

  protected:
    void initSolver(KSP &ksp);

    void initPreconditioner(PC pc);

    void exitPreconditioner(PC pc);

  private:
    int pressureComponent;

    bool pressureNullSpace;

    /// If true, old solution is used for initial guess in solver phase.
    bool useOldInitialGuess;

    /// 0: approximate solve   1: direct solver
    int velocitySolutionMode;

    /// 0: approximate solve   1: direct solver
    int massSolutionMode;

    /// 0: approximate solve   1: direct solver
    int laplaceSolutionMode;

    PetscSolver *massMatrixSolver, *laplaceMatrixSolver, *conDifMatrixSolver;

    NavierStokesSchurData matShellContext;

    double *nu, *invTau;

    SystemVector* solution;

    DOFVector<double>* phase;
    
    IdFct idFct;
  };

}

#endif
