/** \file BoundaryFunctions.h */

#ifndef BOUNDARY_FUNCTIONS_H
#define BOUNDARY_FUNCTIONS_H

#include "AMDiS.h"
#include <time.h>

using namespace std;
using namespace AMDiS;


// parabolic inflow
class ParabolicInflow: public AbstractFunction<double, WorldVector<double> >
{
public:
	ParabolicInflow() : AbstractFunction<double, WorldVector<double> >(2), dim(1.0),c(1.0), dir(1) {};
	ParabolicInflow(double dim_, double c_, int dir_= 0) : AbstractFunction<double, WorldVector<double> >(2), dim(dim_),c(c_),dir(dir_) {};
	double operator()(const WorldVector<double>& x) const  {  
		return c*std::max(0.0, 1.0-sqr(x[1-dir])/sqr(dim)); 
	};
protected: 
	double dim,c;
	int dir;
};

// parabolic inflow
class ParabolicInflowPressure: public AbstractFunction<double, WorldVector<double> >
{
public:
        ParabolicInflowPressure() : AbstractFunction<double, WorldVector<double> >(2), dim(1.0),c(1.0), dir(1), viscosity(1.0) {};
        ParabolicInflowPressure(double dim_, double c_, int dir_= 0, double viscosity_= 1.0) : AbstractFunction<double, WorldVector<double> >(2), dim(dim_),c(c_),dir(dir_),viscosity(viscosity_) {};
        double operator()(const WorldVector<double>& x) const  {  
                return -2.0*c*viscosity*x[dir]/sqr(dim); 
        };
protected: 
        double dim,c,viscosity;
        int dir;
};

class ParabolicInflow_robin: public AbstractFunction<double, WorldVector<double> >
{
public:
	ParabolicInflow_robin() : AbstractFunction<double, WorldVector<double> >(1), y(1.0),c(1.0) {};
	ParabolicInflow_robin(double y_, double c_) : AbstractFunction<double, WorldVector<double> >(1), y(y_),c(c_) {};
	double operator()(const WorldVector<double>& x) const  {  
		double result = c*(1.0-sqr(x[1])/sqr(y)); 
		double n = (x[0]<0.0?-1.0:1.0);
		return -1.0*result*n;
	};
protected: 
	double y,c;
};

// parabolic inflow 2
class BoundaryFct1: public AbstractFunction<double, WorldVector<double> >
{
public:
	BoundaryFct1() : AbstractFunction<double, WorldVector<double> >(1), y(0.205), c(6.0) {};
	BoundaryFct1(double y_, double c_) : AbstractFunction<double, WorldVector<double> >(1), y(y_),c(c_) {};
	double operator()(const WorldVector<double>& x) const  {  
		return 1.0/sqr(2.0*y)*c*(y+x[1])*(y-x[1]);
	};
protected: 
	double y,c;
};

class BoundaryFct1_robin: public AbstractFunction<double, WorldVector<double> >
{
public:
	BoundaryFct1_robin() : AbstractFunction<double, WorldVector<double> >(1), y(0.205), c(6.0) {};
	BoundaryFct1_robin(double y_, double c_) : AbstractFunction<double, WorldVector<double> >(1), y(y_),c(c_) {};
	double operator()(const WorldVector<double>& x) const  {  
		double result = 1.0/sqr(2.0*y)*c*(y+x[1])*(y-x[1]);
		double n = (x[0]<0.0?-1.0:1.0);
		return -1.0*result*n;
	};
protected: 
	double y,c;
};

// time-dependent inflow
class BoundaryFct2: public AbstractFunction<double, WorldVector<double> >, public TimedObject
{
public:
	BoundaryFct2() : AbstractFunction<double, WorldVector<double> >(1), y(0.205), c(6.0) {};
	BoundaryFct2(double y_, double c_) : AbstractFunction<double, WorldVector<double> >(1), y(y_),c(c_) {};
	double operator()(const WorldVector<double>& x) const  {  
		return 1.0/sqr(2.0*y) * sin(M_PI*(*timePtr)/8.0)*c*(y+x[1])*(y-x[1]);
	};
protected: 
	double y,c;
};

class BoundaryFct2_robin: public AbstractFunction<double, WorldVector<double> >, public TimedObject
{
public:
	BoundaryFct2_robin() : AbstractFunction<double, WorldVector<double> >(1), y(0.205), c(6.0) {};
	BoundaryFct2_robin(double y_, double c_) : AbstractFunction<double, WorldVector<double> >(1), y(y_),c(c_) {};
	double operator()(const WorldVector<double>& x) const  {  
		double result = 1.0/sqr(2.0*y) * sin(M_PI*(*timePtr)/8.0)*c*(y+x[1])*(y-x[1]);
		double n = (x[0]<0.0?-1.0:1.0);
		return -1.0*result*n;
	};
protected: 
	double y,c;
};

// Stokes flow (analytic definition)
class StokesFlow : public AbstractFunction<WorldVector<double>, WorldVector<double> >
{
  public:
    StokesFlow(double peclet_) : AbstractFunction<WorldVector<double>, WorldVector<double> >(6), peclet(peclet_) {
      // parameters for pfc
      double R=0.5, density=-0.3, r=0.1, D=1.0;
      Initfile::get("user parameter->colloidal radius",R, 1);
      Initfile::get("pfc-parameter->density",density, 1);
      Initfile::get("pfc-parameter->r",r, 1);
      radius = R/(1.0-R) * 2.0*M_PI/sqrt(3.0);
      D = radius + 2.0*M_PI/sqrt(3.0);

      center.set(0.0);
      Initfile::get("user parameter->colloids position[0]",center, 1);

      c.set(0.0);
      c[0] = peclet/D*abs(r/density);
    }
    WorldVector<double> operator()(const WorldVector<double> &x) const {
      WorldVector<double> xTrans = x-center;
      WorldVector<double> result;

      if(norm(xTrans)<radius)
	      result.set(0.0);
      else {
	      result = c-(3.0*radius/(4.0*norm(xTrans)))*(1.0+sqr(radius)/(3.0*sqr(norm(xTrans))))*c;
	      result+= 3*radius/(4.0*norm(xTrans)*sqr(norm(xTrans)))*(xTrans*c)*(sqr(radius)/sqr(norm(xTrans))-1.0)*xTrans;
      }

      return result;
    }
  protected:
    double peclet;
    double radius;
    WorldVector<double> center,c;
};
class StokesFlow_component : public AbstractFunction<double, WorldVector<double> >
{
  public:
    StokesFlow_component(double peclet_, int comp_) : AbstractFunction<double, WorldVector<double> >(6), peclet(peclet_),comp(comp_) {
      // parameters for pfc
      double R=0.5, density=-0.3, r=0.1, D=1.0;
      Initfile::get("user parameter->colloidal radius",R, 1);
      Initfile::get("pfc-parameter->density",density, 1);
      Initfile::get("pfc-parameter->r",r, 1);
      radius = R/(1.0-R) * 2.0*M_PI/sqrt(3.0);
      D = radius + 2.0*M_PI/sqrt(3.0);

      center.set(0.0);
      Initfile::get("user parameter->colloids position[0]",center, 1);

      c.set(0.0);
      c[0] = peclet/D*abs(r/density);
    }
    double operator()(const WorldVector<double> &x) const {
      WorldVector<double> xTrans = x-center;
      WorldVector<double> result;

      if(norm(xTrans)<radius)
	      result.set(0.0);
      else {
	      result = c-(3.0*radius/(4.0*norm(xTrans)))*(1.0+sqr(radius)/(3.0*sqr(norm(xTrans))))*c;
	      result+= 3*radius/(4.0*norm(xTrans)*sqr(norm(xTrans)))*(xTrans*c)*(sqr(radius)/sqr(norm(xTrans))-1.0)*xTrans;
      }

      return result[comp];
    }
  protected:
    double peclet;
    int comp;
    double radius;
    WorldVector<double> center,c;
};
class StokesFlow_robin : public AbstractFunction<double, WorldVector<double> >
{
  public:
    StokesFlow_robin(double y_, double peclet_) : AbstractFunction<double, WorldVector<double> >(6), y(y_), peclet(peclet_) {
      // parameters for pfc
      double R=0.5, density=-0.3, r=0.1, D=1.0;
      Initfile::get("user parameter->colloidal radius",R, 1);
      Initfile::get("pfc-parameter->density",density, 1);
      Initfile::get("pfc-parameter->r",r, 1);
      radius = R/(1.0-R) * 2.0*M_PI/sqrt(3.0);
      D = radius + 2.0*M_PI/sqrt(3.0);

      center.set(0.0);
      Initfile::get("user parameter->colloids position[0]",center, 1);

      c.set(0.0);
      c[0] = peclet/D*abs(r/density);
    }
    double operator()(const WorldVector<double> &x) const {
      WorldVector<double> xTrans = x-center;
      WorldVector<double> result;

      if(norm(xTrans)<radius)
	      result.set(0.0);
      else {
	      result = c-(3.0*radius/(4.0*norm(xTrans)))*(1.0+sqr(radius)/(3.0*sqr(norm(xTrans))))*c;
	      result+= 3*radius/(4.0*norm(xTrans)*sqr(norm(xTrans)))*(xTrans*c)*(sqr(radius)/sqr(norm(xTrans))-1.0)*xTrans;
      }

      WorldVector<double> n;
      n[0] = (abs(x[1])<y-1.e-8?(x[0]<0.0?-1.0:1.0):0.0);
      n[1] = (abs(x[1])>y-1.e-8?(x[1]<0.0?-1.0:1.0):0.0);
      return -1.0*n*result;
    }
  protected:
    double y,peclet;
    double radius;
    WorldVector<double> center,c;
};

// constant inflow
class ConstantFct : public AbstractFunction<double, WorldVector<double> >
{
public:
  ConstantFct(double c)
    : AbstractFunction<double, WorldVector<double> >(1), constant(c)  {};
  double operator()(const WorldVector<double>& x) const  { return constant;  }

protected:
  double constant;
};

// constant inflow
class ConstantTimeFct : public AbstractFunction<double, WorldVector<double> >, public TimedObject
{
public:
  ConstantTimeFct(double c, double endTime_)
    : AbstractFunction<double, WorldVector<double> >(1), constant(c), endTime(endTime_)  {};
  double operator()(const WorldVector<double>& x) const  { return constant*max(0.0, std::min(1.0, (endTime-(*timePtr))/endTime));  }

protected:
  double constant;
  double endTime;
};
#endif // BOUNDARY_FUNCTIONS_H
