#include "NavierStokes_TaylorHood_RB.h"

using namespace std;
using namespace AMDiS;

NavierStokes_TaylorHood_RB::NavierStokes_TaylorHood_RB(const std::string &name_) :
  super(name_),
  forceDBC(false),
  initialVelocityIsSet(false),
  laplaceType(0),
  nonLinTerm(2),
  oldTimestep(0.0),
  viscosity(1.0),
  c(0.0),
  minus1(-1.0),
  velocity(NULL)
{  
  // force the homogeniouse dirichlet BC if combination of dirichlet and neumann BC
  // are set and AMDiS can not handle this combination automatically
  Initfile::get(name + "->force dirichlet bc", forceDBC);
  if (forceDBC) {
    int numDirichletPoints = 0;
    Initfile::get("number of dirichlet points", numDirichletPoints);
    dirichletPoints.resize(numDirichletPoints);
    for (int i = 0; i < numDirichletPoints; i++)
      Initfile::get("dirichlet point[" + Helpers::toString(i) + "]",dirichletPoints[i]);
  }

  // parameters for navier-stokes
  Initfile::get(name + "->viscosity", viscosity);
  // type of laplace operator: 0... div(nu*grad(u)), 1... div(0.5*nu*(grad(u)+grad(u)^T))
  Initfile::get(name + "->laplace operator", laplaceType); 
  // type of non-linear term: 0... u^old*grad(u_i^old), 1... u'*grad(u_i^old), 2... u^old*grad(u'_i)
  Initfile::get(name + "->non-linear term", nonLinTerm); 

  force.set(0.0);
  Initfile::get(name + "->force", force);

  bool scaleMesh = false;
  Initfile::get("mesh->scale mesh",scaleMesh);
  if (scaleMesh) {
    Initfile::get("mesh->scale",dimension);
  } else
    dimension.set(1.0);
}


NavierStokes_TaylorHood_RB::~NavierStokes_TaylorHood_RB()
{ FUNCNAME("NavierStokes_TaylorHood_RB::~NavierStokes_TaylorHood_RB()");

  if (initialVelocityIsSet) {
    delete initialX;
    delete initialY;
  }
  if (velocity != NULL)
    delete velocity;
  delete fileWriter;
}


void NavierStokes_TaylorHood_RB::initData()
{ FUNCNAME("NavierStokes_TaylorHood_RB::initTimeInterface()");

  if (velocity == NULL)
    velocity = new DOFVector<WorldVector<double> >(getFeSpace(0), "velocity");
  fileWriter = new FileVectorWriter(name + "->velocity->output", getFeSpace()->getMesh(), velocity);

  super::initData();
}


void NavierStokes_TaylorHood_RB::solveInitialProblem(AdaptInfo *adaptInfo)
{ FUNCNAME("NavierStokes_TaylorHood_RB::solveInitialProblem()");

  int initialVelocity = 0, flowDirection = 1;
  Initfile::get(name + "->initial velocity", initialVelocity);
  Initfile::get(name + "->initial velocity->value", c);
  Initfile::get(name + "->initial velocity->flow direction", flowDirection);

  std::vector<AbstractFunction<double, WorldVector<double> >*> initialFcts(2);
  if (initialVelocity == 0) { // no initial velocity
    initialFcts[0]= new ConstantFct(0.0);
    initialFcts[1]= new ConstantFct(0.0);
  } else if(initialVelocity == 1) { // constant inflow
      initialFcts[flowDirection] = new ConstantFct(c);
      initialFcts[1-flowDirection] = new ConstantFct(0.0);
  } else if(initialVelocity == 2) { // parabolic inflow
    double dist = -1.0;
    Initfile::get(name + "->initial velocity->parabolic inflow->boundary dist", dist);
    if (dist <= 0.0)
      dist = 2.0*dimension[1-flowDirection];
    initialFcts[flowDirection] = new ParabolicInflow(dist/2.0, c, flowDirection);
    initialFcts[1-flowDirection] = new ConstantFct(0.0);
  } else
    throw(std::runtime_error("Unknown initial velocity."));

  initialX = initialFcts[0];
  initialY = initialFcts[1];
  initialVelocityIsSet = true;
  
  for (unsigned i = 0; i < dow; ++i) {
    prob->setExactSolutionFct(initialFcts[i], i);
  }
}


void NavierStokes_TaylorHood_RB::transferInitialSolution(AdaptInfo *adaptInfo)
{ FUNCNAME("NavierStokes_TaylorHood_RB::transferInitialSolution()");

  MSG("solve initial problem...\n");
  for (unsigned i = 0; i < dow; ++i) {
    prob->getSolution()->getDOFVector(i)->interpol(prob->getExactSolutionFct(i));
  }
    
  if (dow == 1)
    transformDOF(prob->getSolution()->getDOFVector(0), velocity, new AMDiS::Vec1WorldVec<double>);
  else if (dow == 2)
    transformDOF(prob->getSolution()->getDOFVector(0), prob->getSolution()->getDOFVector(1), velocity, new AMDiS::Vec2WorldVec<double>);
  else if (dow == 3)
    transformDOF(prob->getSolution()->getDOFVector(0), prob->getSolution()->getDOFVector(1), prob->getSolution()->getDOFVector(2), velocity, new AMDiS::Vec3WorldVec<double>);

  for (int i = 0; i < dow; i++)
    prob->setExactSolution(prob->getSolution()->getDOFVector(i), i);

  fileWriter->writeFiles(adaptInfo, false);
  writeFiles(adaptInfo, false);

  // initial parameters for detecting mesh changes
  oldMeshChangeIdx= getMesh()->getChangeIndex();
}


void NavierStokes_TaylorHood_RB::fillOperators()
{ FUNCNAME("NavierStokes_TaylorHood_RB::fillOperators()");
  MSG("NavierStokes_TaylorHood_RB::fillOperators()\n");

  WorldVector<DOFVector<double>*> stage_velocity;
  WorldVector<DOFVector<double>*> un_velocity;
  for (unsigned i = 0; i < dow; ++i) {
    stage_velocity[i] = prob->getStageSolution(i);
    un_velocity[i] = prob->getUnVec(i);
  }
  
  // fill operators for prob
  for (unsigned i = 0; i < dow; ++i) {
    /// < d_t(u_i) , psi >
    prob->addTimeOperator(i, i);
 
    /// < u*grad(u_i) , psi >
    Operator *opUGradU = new Operator(getFeSpace(i), getFeSpace(i));
    opUGradU->addTerm(new WorldVec_FOT(stage_velocity, -1.0), GRD_PHI);
    opUGradU->setUhOld(stage_velocity[i]);
    prob->addVectorOperator(*opUGradU, i);

    Operator *opUGradV = new Operator(getFeSpace(i), getFeSpace(i));
    opUGradV->addTerm(new WorldVec_FOT(un_velocity, -1.0), GRD_PHI);
    prob->addMatrixOperator(*opUGradV, i, i, &minus1, &minus1);

    for (unsigned j = 0; j < dow; ++j) {
      Operator *opVGradU = new Operator(getFeSpace(i), getFeSpace(j));
      opVGradU->addTerm(new PartialDerivative_ZOT(un_velocity[i], j, -1.0));
      prob->addMatrixOperator(*opVGradU, i, j, &minus1, &minus1);
    }

    /// Diffusion-Operator (including Stress-Tensor for space-dependent viscosity)
    addLaplaceTerm(i);
  
    /// < p , d_i(psi) >
    Operator *opGradP = new Operator(getFeSpace(i),getFeSpace(dow));
    opGradP->addTerm(new PartialDerivative_FOT(i, -1.0), GRD_PSI);
    opGradP->setUhOld(prob->getStageSolution(dow));
    prob->addVectorOperator(*opGradP, i);
    prob->addMatrixOperator(*opGradP, i, dow, &minus1, &minus1);
  
    /// external force, i.e. gravitational force
    if (norm(force) > DBL_TOL) {
      Operator *opForce = new Operator(getFeSpace(i), getFeSpace(i));
      opForce->addTerm(new Simple_ZOT(force[i]));
      prob->addVectorOperator(*opForce, i);
    }
  }

  /// div(u) = 0
  for (unsigned i = 0; i < dow; ++i) {
    /// < d_i(u'_i) , psi >
    Operator *opDivU = new Operator(getFeSpace(dow),getFeSpace(i));
    opDivU->addTerm(new PartialDerivative_FOT(i), GRD_PHI);
    opDivU->setUhOld(stage_velocity[i]);
    prob->addVectorOperator(*opDivU, dow);
    prob->addMatrixOperator(*opDivU, dow, i, &minus1, &minus1);
  }
}


void NavierStokes_TaylorHood_RB::addLaplaceTerm(int i)
{ FUNCNAME("NavierStokes_TaylorHood_RB::addLaplaceTerm()");
  MSG("NavierStokes_TaylorHood_RB::addLaplaceTerm(%d)\n",i);

    /// < alpha*[grad(u)+grad(u)^t] , grad(psi) >
    if (laplaceType == 1) {
      for (unsigned j = 0; j < dow; ++j) {
        Operator *opLaplaceUi1 = new Operator(getFeSpace(i), getFeSpace(j));
        opLaplaceUi1->addTerm(new MatrixIJ_SOT(1-i, 1-j, viscosity));
	opLaplaceUi1->setUhOld(prob->getStageSolution(j));
	prob->addVectorOperator(*opLaplaceUi1, i, &minus1, &minus1);
        prob->addMatrixOperator(*opLaplaceUi1, i, j);
      }
    }
    
    /// < alpha*grad(u'_i) , grad(psi) >
    Operator *opLaplaceUi = new Operator(getFeSpace(i), getFeSpace(i));
      opLaplaceUi->addTerm(new Simple_SOT(viscosity));
    opLaplaceUi->setUhOld(prob->getStageSolution(i));
    prob->addVectorOperator(*opLaplaceUi, i, &minus1, &minus1);
    prob->addMatrixOperator(*opLaplaceUi, i, i);
}


void NavierStokes_TaylorHood_RB::fillBoundaryConditions()
{ FUNCNAME("NavierStokes_TaylorHood_RB::fillBoundaryConditions()");

  ERROR("You have to implement some boundary conditions!\n");

//   DOFVector<double> *zeroDOF = new DOFVector<double>(prob->getFeSpace(0), "zero");
//   zeroDOF->set(0.0);
// 
//   /// at rigid wall: no-slip boundary condition
//   for (unsigned j = 0; j < dow; ++j) {
//     prob->addDirichletBC(LINKS, j, j, boundaryDOF[j]);
//     prob->addDirichletBC(RECHTS, j, j, boundaryDOF[j]);
//   }
//   
//   /// at inflow: prescribed velocity
//   for (unsigned j = 0; j < dow; ++j) {
//    prob->addDirichletBC(OBEN, j, j, boundaryDOF[j]);
// //    prob->addDirichletBC(OBEN, j, j, prob->getExactSolution(j)); /// test: Dirichlet-BC at outflow boundary
//   }
// 
//   for (unsigned i = 0; i < dirichletPoints.size(); ++i) {
//     for (unsigned row = 0; row < dow; row++) {
//       prob->addSingularDirichletBC(dirichletPoints[i], row, row, 0.0);
//     }
//   }
}


void NavierStokes_TaylorHood_RB::initTimestep(AdaptInfo *adaptInfo)
{ FUNCNAME("NavierStokes_TaylorHood_RB::initTimestep()");

}


void NavierStokes_TaylorHood_RB::closeTimestep(AdaptInfo *adaptInfo)
{ FUNCNAME("NavierStokes_TaylorHood_RB::closeTimestep()");

  if (dow == 1)
    transformDOF(prob->getSolution()->getDOFVector(0), velocity, new AMDiS::Vec1WorldVec<double>);
  else if (dow == 2)
    transformDOF(prob->getSolution()->getDOFVector(0), prob->getSolution()->getDOFVector(1), velocity, new AMDiS::Vec2WorldVec<double>);
  else if (dow == 3)
    transformDOF(prob->getSolution()->getDOFVector(0), prob->getSolution()->getDOFVector(1), prob->getSolution()->getDOFVector(2), velocity, new AMDiS::Vec3WorldVec<double>);

  fileWriter->writeFiles(adaptInfo, false);
  writeFiles(adaptInfo, false);
}
