#include "AMDiS.h"
#include "CahnHilliardNavierStokes_TwoPhase_RB.h"
#include "SignedDistFunctors.h"
#include "PhaseFieldConvert.h"
#include "Refinement.h"
#include "MeshFunction_Level.h"
#include "time/ExtendedRosenbrockAdaptInstationary.h"

#include "boost/date_time/posix_time/posix_time.hpp"

using namespace AMDiS;
using namespace boost::posix_time;

struct DrivenCavityBC : AbstractFunction<double, WorldVector<double> >
{
  double operator()(const WorldVector<double> &x) const {
    double vel = std::max(0.0, 1.0 - 4.0 * sqr(x[0] - 0.5));
	
	return vel;
  }
};

class CHNS_DrivenCavity : public CahnHilliardNavierStokes_TwoPhase_RB
{
public:
  CHNS_DrivenCavity(std::string name_) : CahnHilliardNavierStokes_TwoPhase_RB(name_) {}

  void solveInitialProblem(AdaptInfo* adaptInfo)
  {
    super::solveInitialProblem(adaptInfo);
    /// horizontale Linie
    double a= 0.0, dir= -1.0;
    double initialEps = eps;
    Initfile::get(name + "->initial epsilon", initialEps);
    Initfile::get(name + "->line->pos", a);
    Initfile::get(name + "->line->direction", dir);
    
    /// create phase-field from signed-dist-function
    if (doubleWell == 0) {
      prob->getSolution()->getDOFVector(0)->interpol(new SignedDistFctToPhaseField(initialEps, new Plane(a, dir)));
    } else {
      prob->getSolution()->getDOFVector(0)->interpol(new Plane(a, dir)); 
      transformDOF(prob->getSolution()->getDOFVector(0),
        new SignedDistToCh(initialEps));
    }
  }
  
  void fillBoundaryConditions()
  { FUNCNAME("NS_DrivenCavity::fillBoundaryConditions()");
    
    AbstractFunction<double, WorldVector<double> > *zero = new AMDiS::Const<double, WorldVector<double> >(0.0);
    size_t dow = Global::getGeo(WORLD);

    /// at rigid wall: no-slip boundary condition
    for (size_t i = 0; i < dow; i++)
      getProblem(0)->addDirichletBC(1, 2+i, 2+i, zero);

	drivenCavityBC = new DrivenCavityBC;
    /// at upper wall: prescribed velocity
    getProblem(0)->addDirichletBC(2, 2, 2, drivenCavityBC);
    getProblem(0)->addDirichletBC(2, 3, 3, zero);
  }
  
private:
  DrivenCavityBC* drivenCavityBC;
};


class RefinementTimeInterface : public CouplingTimeInterface
{
public:

  RefinementTimeInterface(CHNS_DrivenCavity *chnsProb_) :
    chnsProb(chnsProb_),
    refFunction(NULL),
    refinement(NULL)
  {
    addTimeInterface(chnsProb);
  }

  ~RefinementTimeInterface()
  {
    if (refFunction)
      delete refFunction;
    if (refinement)
      delete refinement;
  }

  virtual void initTimeInterface()
  {
    chnsProb->initTimeInterface();
  }

  /// Called at the end of each timestep.
  virtual void closeTimestep(AdaptInfo *adaptInfo)
  {
    CouplingTimeInterface::closeTimestep(adaptInfo);

    refinement->refine(1);
  }

  /// Set initial condition and perform initial refinement
  virtual void solveInitialProblem(AdaptInfo *adaptInfo)
  {
    refFunction= new PhaseFieldRefinement(chnsProb->getMesh());
    refinement= new RefinementLevelDOF(
      chnsProb->getProblem(0)->getFeSpace(0),
      refFunction,
      new PhaseDOFView<double>(chnsProb->getProblem(0)->getSolution()->getDOFVector(0)));

    // initial refinement
    refinement->refine(0);

    bool initialRefinement = true;
    Parameters::get(chnsProb->getName() + "->initial refinement", initialRefinement);
    
    if (initialRefinement) {
      // refine until interfaces is solved
      for (int i = 0; i < 3; ++i) {
	chnsProb->solveInitialProblem(adaptInfo);
	refinement->refine((i < 4 ? 4 : 10));
      }
    }

    CouplingTimeInterface::solveInitialProblem(adaptInfo);
  }

protected:

  CHNS_DrivenCavity *chnsProb;
  PhaseFieldRefinement* refFunction;
  RefinementLevelDOF *refinement;
};


int main(int argc, char** argv)
{ FUNCNAME("main");

  AMDiS::init(argc, argv);

  CHNS_DrivenCavity chnsProb("chns");
  chnsProb.initialize(INIT_ALL);

  RefinementTimeInterface timeInterface(&chnsProb);

  // Adapt-Infos
  AdaptInfo adaptInfo("adapt", chnsProb.getNumComponents());
  ExtendedRosenbrockAdaptInstationary<CHNS_DrivenCavity> adaptInstat("adapt", chnsProb, adaptInfo, timeInterface, adaptInfo);


  ptime start_time = microsec_clock::local_time();
  chnsProb.initTimeInterface();
  int error_code = adaptInstat.adapt(); 
  time_duration td = microsec_clock::local_time()-start_time;

  MSG("elapsed time= %d sec\n", td.total_seconds());

  AMDiS::finalize();

  return error_code;
};
